package com.hz.pm.api.meeting.task;

import cn.hutool.core.util.ArrayUtil;
import com.baomidou.mybatisplus.core.conditions.Wrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.hz.pm.api.common.util.SpringContextHolder;
import com.hz.pm.api.expert.entity.ExpertUserFullInfo;
import com.hz.pm.api.meeting.builder.ExpertInviteBuilder;
import com.hz.pm.api.meeting.entity.domain.Meeting;
import com.hz.pm.api.meeting.entity.domain.MeetingExpert;
import com.hz.pm.api.meeting.entity.dto.*;
import com.hz.pm.api.meeting.entity.enumeration.ExpertAttendStatusEnum;
import com.hz.pm.api.meeting.entity.enumeration.MeetingSettingsTypeEnum;
import com.hz.pm.api.meeting.entity.req.ConfirmedRosterReq;
import com.hz.pm.api.meeting.entity.req.MeetingSettingsSaveReq;
import com.hz.pm.api.meeting.helper.MeetingNotifyHelper;
import com.hz.pm.api.meeting.manage.ExpertInviteManage;
import com.hz.pm.api.meeting.manage.MeetingManage;
import com.hz.pm.api.meeting.manage.MeetingSettingsManage;
import com.hz.pm.api.meeting.service.IExpertInviteAvoidRuleService;
import com.hz.pm.api.meeting.service.IExpertInviteRuleService;
import com.hz.pm.api.meeting.service.IMeetingExpertService;
import com.hz.pm.api.meeting.service.IMeetingService;
import com.hz.pm.api.sms.utils.DateUtil;
import com.ningdatech.basic.util.CollUtils;
import com.ningdatech.basic.util.SpringUtils;
import com.ningdatech.cache.model.cache.CacheHashKey;
import com.ningdatech.cache.repository.CachePlusOps;
import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.springframework.scheduling.concurrent.ThreadPoolTaskScheduler;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.temporal.ChronoUnit;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;


/**
 * <p>
 * 专家随机邀请任务
 * </p>
 *
 * @author WendyYang
 * @since 21:25 2022/8/15
 */
@Slf4j
@Component
@AllArgsConstructor
public class ExpertRandomInviteTask {

    private final RandomInviteProperties properties;
    private static final String MEETING_ID_INVITE_RANDOM = "MEETING_ID_INVITE_RANDOM";

    private static final Duration EXPIRE_TIME = Duration.ofDays(60);

    private final CachePlusOps cachePlusOps;
    @Resource(name = "expertInviteScheduler")
    private ThreadPoolTaskScheduler scheduler;
    private final IMeetingExpertService meetingExpertService;
    private final IExpertInviteRuleService inviteRuleService;
    private final IMeetingService meetingService;
    private final ExpertInviteManage expertInviteManage;
    private final IExpertInviteAvoidRuleService avoidRuleService;
    private final MeetingNotifyHelper meetingNotifyHelper;
    private final MeetingSettingsManage meetingSettingsManage;

    /**
     * 用来存入线程执行句柄, 停止定时任务时使用
     */
    private static final ConcurrentMap<Long, ScheduledFuture<?>> INVITE_TASK_MAP = new ConcurrentHashMap<>();

    //==================================================================================================================

    @PostConstruct
    public void initTask() {
        if (Boolean.FALSE.equals(properties.getEnable())) {
            log.warn("随机邀请已关闭");
            return;
        }
        initInviteTaskAfterStarted();
    }

    /**
     * 构建会议缓存key
     *
     * @param meetingId 会议ID
     * @return 缓存key
     */
    @SuppressWarnings("all")
    private static CacheHashKey cacheKey(Long meetingId) {
        String field = meetingId == null ? null : meetingId.toString();
        return new CacheHashKey(MEETING_ID_INVITE_RANDOM, field, EXPIRE_TIME);
    }

    /**
     * 校验是否在专家抽取任务执行时间之内
     *
     * @return \
     */
    private boolean inInviteTimeRange() {
        MeetingSettingsSaveReq settings = meetingSettingsManage.getSettings(MeetingSettingsTypeEnum.INVITE_CALL_IGNORE_TIME);
        LocalTime currTime = LocalTime.now();
        if (settings != null) {
            InviteCallIgnoreTimeDTO content = settings.getSettingsContent(InviteCallIgnoreTimeDTO.class);
            return DateUtil.between(currTime, content.getStartTime(), content.getStopTime());
        } else {
            int currHour = currTime.getHour();
            return currHour >= properties.getInviteStartHour() && currHour < properties.getInviteEndHour();
        }
    }

    /**
     * 项目重启之后重新初始化邀请任务
     */
    private void initInviteTaskAfterStarted() {
        Map<Long, InviteCacheDTO> caches = cachePlusOps.hGetAll(cacheKey(null));
        if (MapUtils.isEmpty(caches)) {
            log.info("暂无需要初始化的抽取会议信息");
            return;
        }
        Integer inviteDelay = properties.getInviteDelay();
        for (InviteCacheDTO cache : caches.values()) {
            Boolean reInvite = cache.getReInvite();
            LocalDateTime tsTime = cache.getTaskStartTime();
            boolean added = addInviteTask(cache.getMeetingId(), true, inviteDelay, reInvite, tsTime);
            if (!added) {
                cachePlusOps.hDel(cacheKey(cache.getMeetingId()));
            }
        }
    }

    /**
     * 专家抽取校验（抽取数量是否不足）
     *
     * @param meetingId 会议ID
     * @return boolean
     * @author WendyYang
     **/
    private boolean inviteCountCheck(Long meetingId) {
        Map<Long, RandomInviteRuleDTO> ruleMap = inviteRuleService.randomRuleByMeetingId(meetingId);
        if (ruleMap.isEmpty()) {
            return Boolean.FALSE;
        }
        Wrapper<MeetingExpert> query = Wrappers.lambdaQuery(MeetingExpert.class)
                .select(MeetingExpert::getRuleId, MeetingExpert::getStatus)
                .in(MeetingExpert::getRuleId, ruleMap.keySet())
                .eq(MeetingExpert::getStatus, ExpertAttendStatusEnum.AGREED.getCode());
        List<MeetingExpert> experts = meetingExpertService.list(query);
        if (experts.isEmpty()) {
            return Boolean.TRUE;
        }
        Map<Long, Long> cntMap = CollUtils.groupCount(experts, MeetingExpert::getRuleId);
        for (Map.Entry<Long, RandomInviteRuleDTO> entry : ruleMap.entrySet()) {
            Long agreeCnt = cntMap.getOrDefault(entry.getKey(), 0L);
            if (agreeCnt < entry.getValue().getCount()) {
                return Boolean.TRUE;
            }
        }
        return Boolean.FALSE;
    }

    /**
     * 唤醒某个会议的抽取任务
     *
     * @param meetingId 会议ID
     * @param reInvite  是否可邀请已拒绝的专家
     * @author WendyYang
     **/
    public void notifyInviteTask(Long meetingId, boolean... reInvite) {
        boolean tmpReInvite = ArrayUtil.isEmpty(reInvite) || reInvite[0];
        if (INVITE_TASK_MAP.containsKey(meetingId)) {
            return;
        }
        LocalDateTime now = LocalDateTime.now();
        if (addInviteTask(meetingId, false, properties.getInviteDelay(), tmpReInvite, now)) {
            log.info("重置会议的随机抽取状态：{}", meetingId);
            Wrapper<Meeting> update = Wrappers.lambdaUpdate(Meeting.class)
                    .set(Meeting::getInviteStatus, false)
                    .eq(Meeting::getId, meetingId);
            meetingService.update(update);
            InviteCacheDTO cacheVal = InviteCacheDTO.of(meetingId, tmpReInvite, now);
            cachePlusOps.hSet(cacheKey(meetingId), cacheVal);
        }
    }

    /**
     * 添加专家抽取校验任务
     *
     * @param meetingId     会议ID
     * @param checked       是否前置校验
     * @param delayTime     延迟执行时间
     * @param reInvite      是否可以邀请被拒绝的专家
     * @param taskStataTime 任务启动时间
     * @return 是否添加任务成功
     * @author WendyYang
     **/
    private boolean addInviteTask(Long meetingId, boolean checked, int delayTime,
                                  boolean reInvite, LocalDateTime taskStataTime) {
        if (checked && !inviteCountCheck(meetingId)) {
            // 如果抽取数量满足直接返回
            return Boolean.FALSE;
        }
        Instant startTime = Instant.now().plus(delayTime, ChronoUnit.MINUTES);
        ScheduledFuture<?> future = scheduler.scheduleAtFixedRate(() -> {
            ExpertRandomInviteTask bean = SpringContextHolder.getBean(ExpertRandomInviteTask.class);
            try {
                // 抽取专家
                bean.invite(meetingId, reInvite, taskStataTime);
            } catch (Exception e) {
                log.error("执行专家邀请任务异常：{}", meetingId, e);
            }
        }, startTime, Duration.ofMinutes(properties.getInviteFixedRate()));
        INVITE_TASK_MAP.putIfAbsent(meetingId, future);
        log.info("添加专家抽取后台任务：{}", meetingId);
        return Boolean.TRUE;
    }

    /**
     * 创建会议时添加抽取任务
     *
     * @param meetingId     会议ID
     * @param taskStartTime 开始时间
     * @author WendyYang
     **/
    public void addInviteTaskByMeetingCreate(Long meetingId, LocalDateTime taskStartTime) {
        Assert.isTrue(properties.getEnable(), "随机邀请已关闭");
        addInviteTask(meetingId, false, properties.getInviteDelay(), false, taskStartTime);
        InviteCacheDTO cacheVal = InviteCacheDTO.of(meetingId, false, taskStartTime);
        cachePlusOps.hSet(cacheKey(meetingId), cacheVal);
    }

    /**
     * 抽取过程
     *
     * @param meetingId 会议ID
     * @param reInvite  是否可以邀请已拒绝的专家
     * @param tsTime    任务开启时间
     */
    @Transactional(rollbackFor = Exception.class)
    public void invite(Long meetingId, Boolean reInvite, LocalDateTime tsTime) {
        log.info("开始进行专家后台抽取：{}", meetingId);
        if (!inInviteTimeRange()) {
            log.warn("不在会议抽取执行时间：{}", meetingId);
            return;
        }
        Meeting meeting = meetingService.getById(meetingId);
        if (Boolean.TRUE.equals(meeting.getInviteStatus())
                || meeting.getStartTime().isBefore(LocalDateTime.now())) {
            log.info("会议已开始停止抽取：{}", meeting);
            killTaskAndDelCacheMeetingId(meetingId);
            return;
        }
        // 随机邀请规则
        Map<Long, RandomInviteRuleDTO> ruleMap = inviteRuleService.randomRuleByMeetingId(meetingId);
        // 回避规则
        AvoidRuleDTO avoidRule = avoidRuleService.getAvoidInfo(meetingId);
        // 还需要抽取的规则数量
        AtomicInteger notIgnoreCnt = new AtomicInteger(ruleMap.size());
        AtomicInteger notSupportCnt = new AtomicInteger(0);
        LocalDateTime msTime = meeting.getStartTime();
        LocalDateTime meTime = meeting.getStartTime();
        ruleMap.forEach((ruleId, rule) -> {
            List<MeetingExpert> newestExperts = meetingExpertService.listExpertLastByMeetingId(meetingId);
            // 统计通知中与同意参加专家数量
            Map<Long, ExpertCntBO> countMap = expertAttendStatusCount(newestExperts);
            ExpertCntBO cnt = countMap.getOrDefault(ruleId, ExpertCntBO.zeroInit());
            int wouldAttendCnt = cnt.getAgreeCnt() + cnt.getNoticeCnt();
            if (wouldAttendCnt == rule.getCount()) {
                if (cnt.getAgreeCnt().equals(rule.getCount())) {
                    notIgnoreCnt.decrementAndGet();
                }
                return;
            }
            int needInviteCnt = rule.getCount() - wouldAttendCnt;
            // 查询可抽取的专家
            ExpertChoseDTO choseByRule = expertInviteManage.expertReplaceByRandomRule(avoidRule, rule,
                    newestExperts, needInviteCnt, msTime, meTime, tsTime, reInvite);

            if (choseByRule.notEmpty()) {
                List<MeetingExpert> choseExperts = buildSaveExpert(meetingId, ruleId, choseByRule.getExperts());
                // 短信或电话提醒
                meetingNotifyHelper.notifyExperts(meeting, choseExperts);
                meetingExpertService.saveBatch(choseExperts);
                log.info("会议：{} 抽取专家：{} 名", meetingId, choseExperts.size());
            } else {
                // 抽取人数不够
                notSupportCnt.incrementAndGet();
            }
        });
        if (notIgnoreCnt.get() == 0 || notIgnoreCnt.get() == notSupportCnt.get()) {
            log.info("停止会议随机邀请：{} 未完成抽取规则数量 {} 无可抽取专家规则数量 {}", meetingId, notIgnoreCnt, notSupportCnt);
            if (notIgnoreCnt.get() == notSupportCnt.get() && notIgnoreCnt.get() > 0) {
                // 当未完成抽取且无专家可抽取时
                meetingNotifyHelper.sendInviteStopMsg(meeting.getCreateBy(), meetingId, meeting.getName());
            }
            cancelByMeetingIdAndKillTask(meetingId);
        }
        // 所有抽取规则抽取人数满足 自动召开会议
        if (notIgnoreCnt.get() == 0 && notSupportCnt.get() == 0) {
            MeetingManage meetingManage = SpringUtils.getBean(MeetingManage.class);
            ConfirmedRosterReq req = new ConfirmedRosterReq();
            req.setReconfirmed(false);
            req.setMeetingId(meetingId);
            meetingManage.confirmedRoster(req);
        }
    }

    private static List<MeetingExpert> buildSaveExpert(Long meetingId, Long ruleId, List<ExpertUserFullInfo> expertUsers) {
        return CollUtils.convert(expertUsers, w -> {
            MeetingExpert expert = ExpertInviteBuilder.getExpertByRandom(meetingId, w, ruleId);
            expert.setStatus(ExpertAttendStatusEnum.NOTICING.getCode());
            return expert;
        });
    }

    private void killTaskAndDelCacheMeetingId(Long meetingId) {
        cachePlusOps.hDel(cacheKey(meetingId));
        ScheduledFuture<?> future = INVITE_TASK_MAP.get(meetingId);
        if (future != null) {
            INVITE_TASK_MAP.remove(meetingId);
            if (!future.isCancelled()) {
                future.cancel(true);
            }
        }
    }

    @Transactional(rollbackFor = Exception.class)
    public void cancelByMeetingIdAndKillTask(Long meetingId) {
        log.info("终止专家抽取：{}", meetingId);
        meetingService.stopRandomInvite(meetingId);
        killTaskAndDelCacheMeetingId(meetingId);
    }

    //==================================================================================================================

    private static Map<Long, ExpertCntBO> expertAttendStatusCount(List<MeetingExpert> experts) {
        return experts.stream().collect(Collectors.groupingBy(MeetingExpert::getRuleId,
                Collectors.collectingAndThen(Collectors.toList(), w -> {
                    ExpertCntBO cnt = ExpertCntBO.zeroInit();
                    for (MeetingExpert expert : w) {
                        if (ExpertAttendStatusEnum.AGREED.eq(expert.getStatus())) {
                            cnt.incrAgreeCnt();
                        } else if (ExpertAttendStatusEnum.NOTICING.eq(expert.getStatus())) {
                            cnt.incrNoticeCnt();
                        }
                    }
                    return cnt;
                })));
    }

    @Data
    public static class ExpertCntBO {

        private Integer agreeCnt;

        private Integer noticeCnt;

        public ExpertCntBO(Integer agreeCnt, Integer noticeCnt) {
            this.agreeCnt = agreeCnt;
            this.noticeCnt = noticeCnt;
        }

        public void incrAgreeCnt() {
            agreeCnt++;
        }

        public void incrNoticeCnt() {
            noticeCnt++;
        }

        public static ExpertCntBO zeroInit() {
            return new ExpertCntBO(0, 0);
        }

    }

}
